%
% Matlab data parsing and plotting routine
% for Wisepla Biosensor for Tmote Sky v.1.1
%
% Author: Ville-Pekka Seppä and Victor Shia
% Date: 26 Mar 2008
%

function dataOut = plotMoteData_forGUI(hObject, handles, ...
  usbPort, numOfMotes, gcfHandle, dataType, fileName)

if strcmp(dataType, 'realTime')
    s = serial(usbPort,'BaudRate',115200);
    s.InputBufferSize = 8192;
    fopen(s);
    % Empty the serial input buffer
    if s.BytesAvailable
        fread(s, s.BytesAvailable, 'uchar');
    end
    currentRow = 0;
    minRows    = 1;
    
    offsetAccAmt  = -1 * 2^15;
    offsetGyroAmt = -1 * 2^11;
elseif strcmp(dataType, 'parsed')
    load(fileName);
    numOfMotes = length(WearableData.Reading);
    % For parsed data, do not change the data.
    dataOut   = WearableData.Reading;
    
    startDataPt    = str2num(get(handles.startSave,'String'));
    endDataPt      = str2num(get(handles.endSave,'String'));
    sampleCount = size(dataOut{1},1);
    if endDataPt==0
        endDataPt = sampleCount;
        set(handles.endSave,'String',sampleCount);
    end
    if startDataPt==0
        startDataPt = 1;
        set(handles.startSave,'String',1);
    end
    if endDataPt>sampleCount
        warning('End point value is larger than the current data value. Cannot save!');
        return;
    elseif startDataPt>endDataPt
        warning('End point value is smaller than the start point value. Cannot save!');
        return;
    end
    for moteIndex=1:numOfMotes
        moteData{moteIndex} = dataOut{moteIndex}(startDataPt:endDataPt,:);
    end  
    
    minRows    = inf;
    
    for x=1:numOfMotes
        if length(moteData{x}) <= minRows
            minRows = length(moteData{x});
        end
    end
    currentRow = 1;
    
    offsetAccAmt  = 0;
    offsetGyroAmt = 0;
    
elseif strcmp(dataType, 'raw')
    s          = fileName;
    s          = fopen(s);

    currentRow = 0;
    minRows    = 1;
    
    offsetAccAmt  = -1 * 2^15;
    offsetGyroAmt = -1 * 2^11;
end

 
packetHeaderStr = [ ...
  hex2dec('00') ...
  hex2dec('FF') ...
  hex2dec('FF') ...
  hex2dec('FF') ...
  hex2dec('FF') ...
  hex2dec('12') ...
  hex2dec('00') ...
  hex2dec('06') ...
  hex2dec('00') ...
  hex2dec('00') ...
  hex2dec('00') ...
  hex2dec('64') ...
  ];

packetHeaderLen           = length(packetHeaderStr);
packetPayloadLen          = 14;
chunkSize                 = numOfMotes * 256; %512;
dataInputThreshold        = 10; % samples
plotWindowLen             = 800;
%samplingRate              = 20; % Hz
%lostPackets               = 0;
%counter                   = -100; % -100 means counter has not yet been set
chunk(1:chunkSize)        = 0;
%counterOffset(1:numOfMotes) = -1;
%prevCounter(1:numOfMotes)   = -1;
%firstMoteOn               = -1;
numSensors                = 5;
maxWindowCount            = 10240;
lapThreshold              = 100;  
%colorArray                = ['r','b','y','g','k'];

% data to keep track of the number of packets lost
packetsLost = 0;
totalPackets = 0;

motesDataPerChunk         = floor(chunk / numOfMotes);

axes(handles.mote0);
plot([1:plotWindowLen],inf);
axes(handles.mote1);
plot([1:plotWindowLen],inf);
axes(handles.mote2);
plot([1:plotWindowLen],inf);
axes(handles.mote3);
plot([1:plotWindowLen],inf);
axes(handles.mote4);
plot([1:plotWindowLen],inf);

% These are needed for accessing datas in the subplots of the figure
hAxes = gcfHandle.Children;

bufSizeThreshold = 1024; % Aftanner exceeding this amount of packets
                         % raw signal data is moved from 'data'
                         % to 'dataStorage'
dataStoragePtr(1:numOfMotes) = 1;
serialBuf = [];

sensorOffset= [offsetAccAmt offsetAccAmt offsetAccAmt offsetGyroAmt offsetGyroAmt];

% Setting the sensors
moteDataCounter=zeros(1,numOfMotes);
moteWindowCounter=zeros(1,numOfMotes);
dataCounter = 0;
dataSaveCounter = 0;

dataOut = cell(1,numOfMotes);
moteWindow = cell(1, numOfMotes);
for moteIndex=1:numOfMotes
    dataOut{moteIndex} = inf(dataInputThreshold*maxWindowCount, numSensors);
    % Assign physical size of mote window to double the plotWindowLen for
    % buffering possible overflow of extra data
    moteWindow{moteIndex} = zeros(dataInputThreshold*2, numSensors);
end

handles = guidata(hObject);

while ~handles.stopPlotting && currentRow < minRows

  amountOfOldData = length(serialBuf);
  
  % distinguishes between data read from a file and data read from the
  % --- REALTIME MODE ---
  if strcmp(dataType, 'realTime')
      [ chunk, bytesRead, errorMsg ] = fread(s, chunkSize, 'uchar');

      % If there was an error in reading data from USB end
      % this function
      if bytesRead == 0
          %pause(0.02);
          fclose(s);
          return;
      end
      
      % If there was an error in reading data from USB end
      % this function
      if ~strcmp(errorMsg,'')
        disp('Error while reading data from base station:');
        disp(['"' errorMsg '"']);
        disp('Aborting reading...');
        
        break;
      end
  % --- PLAYBACK MODE ---
  elseif strcmp(dataType, 'parsed')
      chunk = [];
      chunkSize = numOfMotes * 256;
      if currentRow == minRows
          break
      end
      %while currentRow <= minRows && length(chunk) < chunkSize
          for j=1:numOfMotes
            tempArray = [];
            temp = moteData{j}(currentRow,:);
            
            % gets rid of bad data
            
            if not (isnan(temp(1)) || temp(1) == inf)% || temp(1) == 0)  if moteID 0 isn't used, uncomment
                for t=1:length(temp)
                    tempArray = [tempArray floor(temp(t)/256) mod(temp(t),256)];
                end
                chunk = [chunk, packetHeaderStr,floor((j-1)/256), mod((j-1),256), floor(currentRow/256), mod(currentRow,256), tempArray, 0]; %  
            end
            
            
          end
          currentRow = currentRow + 1;
      %end
      chunkSize = length(chunk);
      errorMsg = '';
      
  % --- RAW MODE ---
  elseif strcmp(dataType, 'raw')
      tempChunk = [];
      [ tempChunk, bytesRead ] = fread(s, chunkSize * 3, 'uint8=>char');
      tempChunk = tempChunk';
      chunk = [];
      num = 1;
      tempNum = 1;
      % not working yet
      while num <= bytesRead
          tempChunk(num)
          tempChunk(num + 1)
          chunk(tempNum) = hex2dec(tempChunk(num))*16+ hex2dec(tempChunk(num+1));
          tempNum = tempNum + 1;
          num = num + 3;
      end
      
      dataSaveCounter = 1;
  end
  
  % ===================================================================%
  %         Actual Parsing and Graphing of Data                        %
  % ===================================================================%
  
  serialBuf(amountOfOldData + 1: amountOfOldData + chunkSize ) = chunk;
 
  x = 1;

  % This loop extracts the header and payload from the serial stream.
  % Stops analyzing when we are so close to the end of serialBuf chunk
  % that there's no more room for a whole header + payload
  while x < length(serialBuf) - (packetHeaderLen + packetPayloadLen +1)
    % If a header is found starting from cell x in serialBuffer
    if all(serialBuf(x:x+packetHeaderLen-1) == packetHeaderStr)
    
      % THIS PARSING CODE DOESN'T RECOGNIZE ESCAPED 0x7D 0x7E
      % (needs SerialForwarder, doesn't work with raw serial)
      % Store the payload and advance x in buffer
      %             data(dataBufPtr,:) = serialBuf(x + packetHeaderLen: ...
      %                 x + packetHeaderLen+packetPayloadLen - 1);
      %             x = x + packetHeaderLen + packetPayloadLen;
      % end of parsing code

      moteID  = serialBuf(x+packetHeaderLen) * 256 + serialBuf(x+packetHeaderLen + 1) + 1; %once moteID 0 is used
      counter = serialBuf(x+packetHeaderLen + 2) * 256 + serialBuf(x+packetHeaderLen + 3);

      %if isnan(moteID) || moteID == inf || isnan(counter) || counter == inf || moteID == 0
      %    x = x + packetHeaderLen + rawPtr;
      %    break;
      %end
      
      if (moteID > numOfMotes)
          x = x + packetHeaderLen;
          continue;
      end
      
      if moteDataCounter(moteID)==0
          % This is the beginning of sampling
          moteDataCounter(moteID) = counter-1;
          moteWindowCounter(moteID) = 0;
      elseif (counter < moteDataCounter(moteID) + moteWindowCounter(moteID) )
          x = x + packetHeaderLen;
          continue;
      end
      
      % THIS PARSING CODE RECOGNIZES ESCAPED 0x7D 0x7E
      % (works with raw serial data from the mote)
      rawPtr = 5;  % Skip moteID double byte and counter double byte
      parsedPtr = 1;
      data = zeros(1,packetPayloadLen);
      while parsedPtr <= numSensors*2

        data(parsedPtr) = serialBuf(x + packetHeaderLen + rawPtr - 1);

        % Let's skip one byte in the serialBuffer if we encounter
        % an escaped character (0x7D or 0x7E)
        if strcmp(dataType, 'realTime') || strcmp(dataType, 'raw')
            if data( parsedPtr) == 125 || data(parsedPtr) == 126
              rawPtr = rawPtr + 1;
            end
        end
        parsedPtr = parsedPtr + 1;
        rawPtr = rawPtr + 1;
      end
      
      x = x + packetHeaderLen + rawPtr;

      % add data to window
        
      packetLap = counter - (moteDataCounter(moteID) + moteWindowCounter(moteID));
      
      if packetLap>lapThreshold
          % Possibly a corrupted packet
          continue;
      end

      convertedData = zeros(1,numSensors);
      for sensorIndex=0:numSensors-1
          convertedData(sensorIndex+1) = data(2*sensorIndex+1).*256 + data(2*sensorIndex+2);
      end
      convertedData = convertedData + sensorOffset;
      if max(abs(convertedData))>5000
        continue;
      end
      
      if moteWindowCounter(moteID) == 0
          % First signal
          moteWindow{moteID}(1, :) = convertedData;
      else
          % if exists dropping packets, interpolate
          for windowIndex=1: packetLap
              moteWindow{moteID}(moteWindowCounter(moteID) + windowIndex, :) = (1-windowIndex/packetLap)*moteWindow{moteID}(moteWindowCounter(moteID),:) ...
                  + (windowIndex/packetLap)*convertedData;
              %moteWindow{moteID}(moteWindowCounter(moteID) + windowIndex, :) = moteWindow{moteID}(moteWindowCounter(moteID),:);
          end
      end
      moteWindowCounter(moteID) = moteWindowCounter(moteID) + packetLap;
    else
        % If no header was found, move to next byte
        x = x + 1;
    end
  end
  
  % Move the not yet parsed end of previous chunk
  % to the beginning of the buffer to be parsed first
  serialBuf = serialBuf(x:length(serialBuf));
  
  % Put moteWindow to dataOut
  dataWindowLen = max(moteWindowCounter) -1;
  if dataWindowLen >= dataInputThreshold
      if dataCounter + dataWindowLen>plotWindowLen
          invervalStart = dataCounter + dataWindowLen - plotWindowLen;
          invervalEnd = dataCounter + dataWindowLen;
      else
          invervalStart = 1;
          invervalEnd = plotWindowLen;
      end
      
      if handles.saveParsedData == 1&& dataSaveCounter==0
          % assign the start saving point
          dataSaveCounter = dataCounter;
      end
      
      for moteIndex=1:numOfMotes
          if moteWindowCounter(moteIndex)>=dataWindowLen
              % Data in mote window is longer than plot window length, CUT
              dataOut{moteIndex}(dataCounter+1:dataCounter+dataWindowLen,:) = moteWindow{moteIndex}(1:dataWindowLen,:);
              moteWindow{moteIndex}(1:moteWindowCounter(moteIndex)-dataWindowLen,:) = moteWindow{moteIndex}(dataWindowLen+1:moteWindowCounter(moteIndex),:);
              moteWindowCounter(moteIndex) = moteWindowCounter(moteIndex) - dataWindowLen;
              moteDataCounter(moteIndex) = moteDataCounter(moteIndex) + dataWindowLen;
          elseif moteWindowCounter(moteIndex) ~= 0
              % Data in mote window is shorter than plot window length,
              % INTERPOLATE
              for sensorIndex=1:numSensors
                  dataOut{moteIndex}(dataCounter+1:dataCounter+dataWindowLen,sensorIndex) = interpft(moteWindow{moteIndex}(1:moteWindowCounter(moteIndex),sensorIndex),dataWindowLen);
              end
              moteDataCounter(moteIndex) = moteDataCounter(moteIndex) + moteWindowCounter(moteIndex);
              moteWindowCounter(moteIndex) = 0;
          end
      end
      
      axes(handles.mote0);
      plot(invervalStart:invervalEnd,dataOut{1}(invervalStart:invervalEnd,[1 4 5]));
      xlim([invervalStart,invervalEnd]);
      ylim([-3000,3000]);
      axes(handles.mote1);
      plot(invervalStart:invervalEnd,dataOut{2}(invervalStart:invervalEnd,[1 4 5]));
      xlim([invervalStart,invervalEnd]);
      ylim([-3000,3000]);
      axes(handles.mote2);
      plot(invervalStart:invervalEnd,dataOut{3}(invervalStart:invervalEnd,[1 4 5]));
      xlim([invervalStart,invervalEnd]);
      ylim([-3000,3000]);
      axes(handles.mote3);
      plot(invervalStart:invervalEnd,dataOut{4}(invervalStart:invervalEnd,[1 4 5]));
      xlim([invervalStart,invervalEnd]);
      ylim([-3000,3000]);
      axes(handles.mote4);
      plot(invervalStart:invervalEnd,dataOut{5}(invervalStart:invervalEnd,[1 4 5]));
      xlim([invervalStart,invervalEnd]);
      ylim([-3000,3000]);
      
      dataCounter = dataCounter + dataWindowLen;
      handles = guidata(hObject);
      
      if dataCounter>=dataWindowLen*maxWindowCount
          warning('Data exceed storage, STOP!');
          break;
      end
  end
end
% Currently the last mote window will not be saved to dataOut

% if we are reading from a file, close the file
if strcmp(dataType, 'realTime') || strcmp(dataType, 'raw')
    fclose(s);
end

if ~strcmp(dataType, 'parsed')
    if dataSaveCounter==0
        dataOut = [];
    else
        for moteIndex=1:numOfMotes
            dataOut{moteIndex} = dataOut{moteIndex}(dataSaveCounter:dataCounter,:);
        end
    end
end

%disp(['packetsLost  = ' packetsLost]);
%disp(['totalPackets = ' totalPackets]);
%disp(['% lost       = ' packetsLost / (totalPackets + 1)]);

%save = handles.saveParsedData;